@@ -1,24 +1,43 @@
 import ssl
 import asyncio
+import inspect
+import pathlib
 import argparse
 import subprocess
+from typing import List
 
 from aiohttp import web
 
-from dffml.util.cli.arg import Arg
 from dffml.util.cli.cmd import CMD
 from dffml import Model, Sources, BaseSource
 from dffml.util.cli.parser import list_action
 from dffml.util.entrypoint import entrypoint
 from dffml.util.asynchelper import AsyncContextManagerList
+from dffml.base import config, field
 
 from .routes import Routes
 
 
+@config
+class TLSCMDConfig:
+    key: str = field(
+        "Path to key file", default="server.key",
+    )
+    cert: str = field(
+        "Path to cert file", default="server.pem",
+    )
+
+
 class TLSCMD(CMD):
 
-    arg_key = Arg("-key", help="Path to key file", default="server.key")
-    arg_cert = Arg("-cert", help="Path to cert file", default="server.pem")
+    CONFIG = TLSCMDConfig
+
+
+@config
+class CreateTLSServerConfig(TLSCMDConfig):
+    bits: int = field(
+        "Number of bits to use for key", default=4096,
+    )
 
 
 class CreateTLSServer(TLSCMD):
@@ -26,9 +45,7 @@ class CreateTLSServer(TLSCMD):
     Used to generate server key and cert
     """
 
-    arg_bits = Arg(
-        "-bits", help="Number of bits to use for key", default=4096, type=int
-    )
+    CONFIG = CreateTLSServerConfig
 
     async def run(self):
         subprocess.call(
@@ -52,6 +69,28 @@ class CreateTLSServer(TLSCMD):
         )
 
 
+@config
+class CreateTLSClientConfig:
+    bits: int = field(
+        "Number of bits to use for key", default=4096,
+    )
+    key: str = field(
+        "Path to client key file", default="client.key",
+    )
+    cert: str = field(
+        "Path to client cert file", default="client.pem",
+    )
+    csr: str = field(
+        "Path to client csr file", default="client.csr",
+    )
+    server_key: str = field(
+        "Path to server key file", default="server.key",
+    )
+    server_cert: str = field(
+        "Path to server cert file", default="server.pem",
+    )
+
+
 class CreateTLSClient(CMD):
     """
     Create TLS client key and cert (used to authenticate to HTTP API server).
@@ -66,20 +105,7 @@ class CreateTLSClient(CMD):
 
     CLI_FORMATTER_CLASS = argparse.RawDescriptionHelpFormatter
 
-    arg_bits = Arg(
-        "-bits", help="Number of bits to use for key", default=4096, type=int
-    )
-    arg_key = Arg("-key", help="Path to client key file", default="client.key")
-    arg_cert = Arg(
-        "-cert", help="Path to client cert file", default="client.pem"
-    )
-    arg_csr = Arg("-csr", help="Path to client csr file", default="client.csr")
-    arg_server_key = Arg(
-        "-server-key", help="Path to server key file", default="server.key"
-    )
-    arg_server_cert = Arg(
-        "-server-cert", help="Path to server cert file", default="server.pem"
-    )
+    CONFIG = CreateTLSClientConfig
 
     async def run(self):
         subprocess.check_call(
@@ -129,20 +155,107 @@ class CreateTLS(TLSCMD):
     client = CreateTLSClient
 
 
-class MultiCommCMD(CMD):
-
-    arg_mc_config = Arg(
-        "-mc-config",
-        dest="mc_config",
-        default=None,
-        help="MultiComm config directory",
+@config
+class MultiCommCMDConfig:
+    mc_config: str = field(
+        "MultiComm config directory", default=None,
     )
-    arg_mc_atomic = Arg(
-        "-mc-atomic",
-        dest="mc_atomic",
+    mc_atomic: bool = field(
+        "No routes other than dataflows registered at startup",
         action="store_true",
         default=False,
-        help="No routes other than dataflows registered at startup",
+    )
+
+
+class MultiCommCMD(CMD):
+
+    CONFIG = MultiCommCMDConfig
+
+
+class RedirectFormatError(Exception):
+    """
+    Raises when -redirects was not divisible by 3
+    """
+
+
+class ParseRedirectsAction(argparse.Action):
+    def __call__(self, parser, namespace, values, option_string=None):
+        if not isinstance(values, list):
+            values = [values]
+        if len(values) % 3 != 0:
+            raise RedirectFormatError(
+                inspect.cleandoc(
+                    r"""
+                -redirect usage: METHOD SOURCE_PATH DESTINATION_PATH
+
+                Examples
+                --------
+
+                Redirect / to /index.html for GET requests
+
+                    -redirect GET / /index.html
+
+                Redirect / to /index.html for GET requests and /signup to
+                /mysignup for POST requests
+
+                    -redirect \
+                        GET / /index.html \
+                        POST /signup /mysignup
+                """
+                ).strip()
+            )
+        setattr(namespace, self.dest, values)
+
+
+@config
+class ServerConfig(TLSCMDConfig, MultiCommCMDConfig):
+    port: int = field(
+        "Port to bind to", default=8080,
+    )
+    addr: str = field(
+        "Address to bind to", default="127.0.0.1",
+    )
+    upload_dir: str = field(
+        "Directory to store uploaded files in", default=None,
+    )
+    static: str = field(
+        "Directory to serve static content from", default=None,
+    )
+    js: bool = field(
+        "Serve JavaScript API file at /api.js",
+        default=False,
+        action="store_true",
+    )
+    insecure: bool = field(
+        "Start without TLS encryption", action="store_true", default=False,
+    )
+    cors_domains: List[str] = field(
+        "Domains to allow CORS for (see keys in defaults dict for aiohttp_cors.setup)",
+        default_factory=lambda: [],
+    )
+    allow_caching: bool = field(
+        "Allow caching of HTTP responses", action="store_true", default=False,
+    )
+    models: Model = field(
+        "Models configured on start",
+        default_factory=lambda: AsyncContextManagerList(),
+        action=list_action(AsyncContextManagerList),
+        labeled=True,
+    )
+    sources: Sources = field(
+        "Sources configured on start",
+        default_factory=lambda: Sources(),
+        action=list_action(Sources),
+        labeled=True,
+    )
+    redirect: List[str] = field(
+        "list of METHOD SOURCE_PATH DESTINATION_PATH pairs, number of elements must be divisible by 3",
+        action=ParseRedirectsAction,
+        default_factory=lambda: [],
+    )
+    portfile: pathlib.Path = field(
+        "File to write bound port to when starting. Helpful when port 0 was requeseted to bind to any free port",
+        default=None,
     )
 
 
@@ -156,50 +269,7 @@ class Server(TLSCMD, MultiCommCMD, Routes):
     RUN_YIELD_FINISH = False
     INSECURE_NO_TLS = False
 
-    arg_port = Arg("-port", help="Port to bind to", type=int, default=8080)
-    arg_addr = Arg("-addr", help="Address to bind to", default="127.0.0.1")
-    arg_upload_dir = Arg(
-        "-upload-dir",
-        help="Directory to store uploaded files in",
-        default=None,
-    )
-    arg_static = Arg(
-        "-static", help="Directory to serve static content from", default=None
-    )
-    arg_js = Arg(
-        "-js",
-        help="Serve JavaScript API file at /api.js",
-        default=False,
-        action="store_true",
-    )
-    arg_insecure = Arg(
-        "-insecure",
-        help="Start without TLS encryption",
-        action="store_true",
-        default=False,
-    )
-    arg_cors_domains = Arg(
-        "-cors-domains",
-        help="Domains to allow CORS for (see keys in defaults dict for aiohttp_cors.setup)",
-        nargs="+",
-        default=[],
-    )
-    arg_models = Arg(
-        "-models",
-        help="Models configured on start",
-        nargs="+",
-        default=AsyncContextManagerList(),
-        type=Model.load_labeled,
-        action=list_action(AsyncContextManagerList),
-    )
-    arg_sources = Arg(
-        "-sources",
-        help="Sources configured on start",
-        nargs="+",
-        default=Sources(),
-        type=BaseSource.load_labeled,
-        action=list_action(Sources),
-    )
+    CONFIG = ServerConfig
 
     async def start(self):
         if self.insecure:
@@ -207,10 +277,18 @@ class Server(TLSCMD, MultiCommCMD, Routes):
                 self.runner, host=self.addr, port=self.port
             )
         else:
+            cert_path = pathlib.Path(self.cert).resolve()
+            if not cert_path.is_file():
+                raise FileNotFoundError(
+                    f"Server cert not found: {cert_path!s}"
+                )
+            key_path = pathlib.Path(self.key)
+            if not key_path.is_file():
+                raise FileNotFoundError(f"Server key not found: {key_path!s}")
             ssl_context = ssl.create_default_context(
-                purpose=ssl.Purpose.SERVER_AUTH, cafile=self.cert
+                purpose=ssl.Purpose.SERVER_AUTH, cafile=str(cert_path)
             )
-            ssl_context.load_cert_chain(self.cert, self.key)
+            ssl_context.load_cert_chain(str(cert_path), str(key_path))
             self.site = web.TCPSite(
                 self.runner,
                 host=self.addr,
@@ -235,6 +313,9 @@ class Server(TLSCMD, MultiCommCMD, Routes):
             self.mc_atomic = False
             await self.register_directory(self.mc_config)
             self.mc_atomic = atomic
+        # Write out port to file
+        if self.portfile is not None:
+            pathlib.Path(self.portfile).write_text(str(self.port))
         try:
             # If we are testing then RUN_YIELD will be an asyncio.Event
             if self.RUN_YIELD_START is not False:
