@@ -1,20 +1,24 @@
 import os
 import json
+import asyncio
 import secrets
 import inspect
 import pathlib
 import traceback
+import urllib.parse
 import pkg_resources
 from functools import wraps
 from http import HTTPStatus
 from functools import partial
 from dataclasses import dataclass
 from contextlib import AsyncExitStack
-from typing import List, Union, AsyncIterator, Type, NamedTuple, Dict
+from typing import List, Union, AsyncIterator, Type, NamedTuple, Dict, Any
 
-from aiohttp import web
 import aiohttp_cors
+from aiohttp import web
 
+
+from dffml import Sources, MemorySource
 from dffml.record import Record
 from dffml.df.types import DataFlow, Input
 from dffml.df.multicomm import MultiCommInAtomicMode, BaseMultiCommContext
@@ -24,8 +28,9 @@ from dffml.df.memory import (
     MemoryInputSetConfig,
     StringInputSetContext,
 )
-from dffml.base import MissingConfig
 from dffml.model import Model
+from dffml.base import MissingConfig
+from dffml.util.data import traverse_get
 from dffml.source.source import BaseSource, SourcesContext
 from dffml.util.entrypoint import EntrypointNotFound, entrypoint
 
@@ -39,6 +44,12 @@ API_JS_BYTES = pathlib.Path(
 SECRETS_TOKEN_BITS = 384
 SECRETS_TOKEN_BYTES = int(SECRETS_TOKEN_BITS / 8)
 
+# Headers required to set no-cache for confidential web pages
+DISALLOW_CACHING = {
+    "Pragma": "no-cache",
+    "Cache-Control": "no-cache, no-store, max-age=0, must-revalidate",
+    "Expires": "-1",
+}
 
 OK = {"error": None}
 SOURCE_NOT_LOADED = {"error": "Source not loaded"}
@@ -132,10 +143,72 @@ def mctx_route(handler):
 
 
 class HTTPChannelConfig(NamedTuple):
+    """
+    Config for channels.
+
+    Parameters
+    ++++++++++
+        path : str
+            Route in server.
+        dataflow : DataFlow
+            Flow to which inputs from request to path is forwarded too.
+        input_mode : str
+            Mode according to which input data is passed to the dataflow,default:"default".
+                - "default" :
+                    Inputs are expected to be mapping of context to list of input
+                        to definition mappings
+                        eg:'{
+                        "insecure-package":
+                            [
+                                {
+                                    "value":"insecure-package",
+                                    "definition":"package"
+                                }
+                            ]
+                        }'
+                - "preprocess:definition_name" :
+                    Input as whole is treated as value with the given definition.
+                    Supported 'preporcess' tags : [json,text,bytes,stream]
+        output_mode : str
+            Mode according to which output from dataflow is treated.
+                - bytes:content_type:OUTPUT_KEYS :
+                    OUTPUT_KEYS are . seperated string which is used as keys to traverse the
+                    ouput of the flow.
+                    eg:
+                        `results = {
+                            "post_input":
+                                {
+                                    "hex":b'speak'
+                                }
+                        }`
+                    then bytes:post_input.hex will return b'speak'.
+                - text:OUTPUT_KEYS
+                - json
+                    - output of dataflow (Dict) is passes as json
+        immediate_response: Dict[str,Any]
+            If provided with a reponse, server responds immediatly with
+            it, whilst scheduling to run the dataflow.
+            Expected keys:
+                - status: HTTP status code for the response
+                - content_type: MIME type.If not given, determined
+                    from the presence of body/text/json
+                - body/text/json: One of this according to content_type
+                - headers
+            eg:
+                {
+                    "status": 200,
+                    "content_type": "application/json",
+                    "data": {"text": "ok"},
+                }
+    """
+
     path: str
-    presentation: str
-    asynchronous: bool
     dataflow: DataFlow
+    output_mode: str = "json"
+    asynchronous: bool = False
+    input_mode: str = "default"
+    forward_headers: str = None
+    immediate_response: Dict[str, Any] = None
 
     @classmethod
     def _fromdict(cls, **kwargs):
@@ -145,66 +218,191 @@ class HTTPChannelConfig(NamedTuple):
 
 @entrypoint("http")
 class Routes(BaseMultiCommContext):
-    PRESENTATION_OPTIONS = ["json", "blob", "text"]
+    IO_MODES = ["json", "text", "bytes", "stream"]
 
     async def get_registered_handler(self, request):
         return self.app["multicomm_routes"].get(request.path, None)
 
-    async def multicomm_dataflow(self, config, request):
+    async def _multicomm_dataflow(self, config, request):
         # Seed the network with inputs given by caller
         # TODO(p0,security) allowlist of valid definitions to seed (set
         # Input.origin to something other than seed)
         inputs = []
         # If data was sent add those inputs
         if request.method == "POST":
-            # Accept a list of input data
-            # TODO validate that input data is dict of list of inputs each item
-            # has definition and value properties
-            for ctx, client_inputs in (await request.json()).items():
-                for input_data in client_inputs:
-                    if (
-                        not input_data["definition"]
-                        in config.dataflow.definitions
-                    ):
-                        return web.json_response(
-                            {
-                                "error": f"Missing definition for {input_data['definition']} in dataflow"
-                            },
-                            status=HTTPStatus.NOT_FOUND,
+            # Accept a list of input data according to config.input_mode
+            if config.input_mode == "default":
+                # TODO validate that input data is dict of list of inputs each item
+                # has definition and value properties
+                for ctx, client_inputs in (await request.json()).items():
+                    for input_data in client_inputs:
+                        if (
+                            not input_data["definition"]
+                            in config.dataflow.definitions
+                        ):
+                            return web.json_response(
+                                {
+                                    "error": f"Missing definition for {input_data['definition']} in dataflow"
+                                },
+                                status=HTTPStatus.NOT_FOUND,
+                            )
+
+                    inputs.append(
+                        MemoryInputSet(
+                            MemoryInputSetConfig(
+                                ctx=StringInputSetContext(ctx),
+                                inputs=[
+                                    Input(
+                                        value=input_data["value"],
+                                        definition=config.dataflow.definitions[
+                                            input_data["definition"]
+                                        ],
+                                    )
+                                    for input_data in client_inputs
+                                ]
+                                + (
+                                    [
+                                        Input(
+                                            value=request.headers,
+                                            definition=config.dataflow.definitions[
+                                                config.forward_headers
+                                            ],
+                                        )
+                                    ]
+                                    if config.forward_headers
+                                    else []
+                                ),
+                            )
                         )
+                    )
+            elif ":" in config.input_mode:
+                preprocess_mode, *input_def = config.input_mode.split(":")
+                input_def = ":".join(input_def)
+                if input_def not in config.dataflow.definitions:
+                    return web.json_response(
+                        {
+                            "error": f"Missing definition for {input_def} in dataflow"
+                        },
+                        status=HTTPStatus.NOT_FOUND,
+                    )
+
+                if preprocess_mode == "json":
+                    value = await request.json()
+                elif preprocess_mode == "text":
+                    value = await request.text()
+                elif preprocess_mode == "bytes":
+                    value = await request.read()
+                elif preprocess_mode == "stream":
+                    value = request.content
+                else:
+                    return web.json_response(
+                        {
+                            "error": f"preprocess tag must be one of {self.IO_MODES}, got {preprocess_mode}"
+                        },
+                        status=HTTPStatus.NOT_FOUND,
+                    )
+
                 inputs.append(
                     MemoryInputSet(
                         MemoryInputSetConfig(
-                            ctx=StringInputSetContext(ctx),
+                            ctx=StringInputSetContext("post_input"),
                             inputs=[
                                 Input(
-                                    value=input_data["value"],
+                                    value=value,
                                     definition=config.dataflow.definitions[
-                                        input_data["definition"]
+                                        input_def
                                     ],
                                 )
-                                for input_data in client_inputs
-                            ],
+                            ]
+                            + (
+                                [
+                                    Input(
+                                        value=request.headers,
+                                        definition=config.dataflow.definitions[
+                                            config.forward_headers
+                                        ],
+                                    )
+                                ]
+                                if config.forward_headers
+                                else []
+                            ),
                         )
                     )
                 )
+
+            else:
+                raise NotImplementedError(
+                    "Input modes other than default,preprocess:definition_name  not yet implemented"
+                )
+
         # Run the operation in an orchestrator
         # TODO(dfass) Create the orchestrator on startup of the HTTP API itself
-        async with MemoryOrchestrator.basic_config() as orchestrator:
+        async with MemoryOrchestrator() as orchestrator:
             # TODO(dfass) Create octx on dataflow registration
             async with orchestrator(config.dataflow) as octx:
                 results = {
                     str(ctx): result async for ctx, result in octx.run(*inputs)
                 }
-                # TODO Implement input and presentation stages?
-                """
-                if config.presentation == "blob":
-                    return web.Response(body=results)
-                elif config.presentation == "text":
-                    return web.Response(text=results)
+
+                if config.output_mode == "json":
+                    return web.json_response(results)
+
+                # content_info is a List[str] ([content_type,output_keys])
+                # in case of stream,bytes and string in others
+                postprocess_mode, *content_info = config.output_mode.split(":")
+
+                if postprocess_mode == "stream":
+                    # stream:text/plain:get_single.beef
+                    raise NotImplementedError(
+                        "output mode  not yet implemented"
+                    )
+
+                elif postprocess_mode == "bytes":
+                    content_type, output_keys = content_info
+                    output_data = traverse_get(results, output_keys)
+                    return web.Response(body=output_data)
+
+                elif postprocess_mode == "text":
+                    output_data = traverse_get(results, content_info[0])
+                    return web.Response(text=output_data)
+
                 else:
-                """
-                return web.json_response(results)
+                    return web.json_response(
+                        {"error": f"output mode not valid"},
+                        status=HTTPStatus.NOT_FOUND,
+                    )
+
+    async def multicomm_dataflow(self, config, request):
+        if config.immediate_response:
+            asyncio.create_task(self._multicomm_dataflow(config, request))
+            ir = config.immediate_response
+            content_type = None
+            if "content_type" in ir:
+                content_type = ir["content_type"]
+            else:
+                if "data" in ir:
+                    content_type = "application/json"
+                elif "text" in ir:
+                    content_type = "text/plain"
+                elif "body" in ir:
+                    content_type = "application/octet-stream"
+
+            if content_type == "application/json":
+                return web.json_response(
+                    data={} if not "data" in ir else ir["data"],
+                    status=200 if not "status" in ir else ir["status"],
+                    headers=None if not "headers" in ir else ir["headers"],
+                )
+            else:
+                return web.Response(
+                    body=None if not "body" in ir else ir["body"],
+                    text=None if not "text" in ir else ir["text"],
+                    status=200 if not "status" in ir else ir["status"],
+                    headers=None if not "headers" in ir else ir["headers"],
+                    content_type=content_type,
+                )
+        else:
+            return await self._multicomm_dataflow(config, request)
 
     async def multicomm_dataflow_asynchronous(self, config, request):
         # TODO allow list of valid definitions to seed
@@ -237,31 +435,50 @@ class Routes(BaseMultiCommContext):
 
     @web.middleware
     async def error_middleware(self, request, handler):
+        # URL decode match_info values to fix bug where some versions of Python
+        # / aiohttp do not decode them
+        for key, value in request.match_info.items():
+            request.match_info[key] = urllib.parse.unquote(
+                value, errors="strict"
+            )
         try:
             # HACK This checks if aiohttp's builtin not found handler is going
             # to be called
             # Check if handler is the not found handler
-            if "Not Found" in str(handler):
+            if (
+                "Not Found" in str(handler)
+                or "StaticResource" in str(handler)
+                or "Method Not Allowed" in str(handler)
+            ):
                 # Run get_registered_handler to see if we can find the handler
                 new_handler = await self.get_registered_handler(request)
                 if new_handler is not None:
                     handler = new_handler
-            return await handler(request)
+            response = await handler(request)
         except web.HTTPException as error:
             response = {"error": error.reason}
             if error.text is not None:
                 response["error"] = error.text
-            return web.json_response(response, status=error.status)
+            response = web.json_response(response, status=error.status)
         except Exception as error:  #  pragma: no cov
             self.logger.error(
                 "ERROR handling %s: %s",
                 request,
                 traceback.format_exc().strip(),
             )
-            return web.json_response(
+            response = web.json_response(
                 {"error": "Internal Server Error"},
                 status=HTTPStatus.INTERNAL_SERVER_ERROR,
             )
+        # Disable request caching unless allowed explicitly
+        if not self.allow_caching:
+            self.set_no_cache(response)
+        return response
+
+    def set_no_cache(self, response):
+        # Set headers required to set no-cache for confidential web pages
+        for header, value in DISALLOW_CACHING.items():
+            response.headers[header] = value
 
     async def service_upload(self, request):
         if self.upload_dir is None:
@@ -457,10 +674,10 @@ class Routes(BaseMultiCommContext):
     @mcctx_route
     async def multicomm_register(self, request, mcctx):
         config = mcctx.register_config()._fromdict(**(await request.json()))
-        if config.presentation not in self.PRESENTATION_OPTIONS:
+        if config.output_mode not in self.IO_MODES:
             return web.json_response(
                 {
-                    "error": f"{config.presentation!r} is not a valid presentation option: {self.PRESENTATION_OPTIONS!r}"
+                    "error": f"{config.output_mode!r} is not a valid output_mode option: {self.IO_MODES!r}"
                 },
                 status=HTTPStatus.BAD_REQUEST,
             )
@@ -596,25 +813,27 @@ class Routes(BaseMultiCommContext):
                 status=HTTPStatus.BAD_REQUEST,
             )
         # Get the records
-        records: Dict[str, Record] = {
-            key: Record(key, data=record_data)
-            for key, record_data in (await request.json()).items()
-        }
-        # Create an async generator to feed records
-        async def record_gen():
-            for record in records.values():
-                yield record
-
-        # Feed them through prediction
-        return web.json_response(
-            {
-                "iterkey": None,
-                "records": {
-                    record.key: record.export()
-                    async for record in mctx.predict(record_gen())
-                },
-            }
-        )
+        records: Dict[str, Record] = {}
+        # Create a source with will provide the records
+        async with Sources(
+            MemorySource(
+                records=[
+                    Record(key, data=record_data)
+                    for key, record_data in (await request.json()).items()
+                ]
+            )
+        ) as source:
+            async with source() as sctx:
+                # Feed them through prediction
+                return web.json_response(
+                    {
+                        "iterkey": None,
+                        "records": {
+                            record.key: record.export()
+                            async for record in mctx.predict(sctx)
+                        },
+                    }
+                )
 
     async def api_js(self, request):
         return web.Response(
@@ -622,6 +841,15 @@ class Routes(BaseMultiCommContext):
             headers={"Content-Type": "application/javascript"},
         )
 
+    def mkredirect(self, destination_path):
+        async def redirector(request):
+            return web.Response(
+                status=HTTPStatus.TEMPORARY_REDIRECT,
+                headers={"Location": destination_path},
+            )
+
+        return redirector
+
     async def on_shutdown(self, app):
         self.logger.debug("Shutting down service and exiting all contexts")
         await app["exit_stack"].__aexit__(None, None, None)
@@ -740,6 +968,11 @@ class Routes(BaseMultiCommContext):
                 ),
             ]
         )
+        # Redirects
+        for method, src, dst in (
+            self.redirect[i : i + 3] for i in range(0, len(self.redirect), 3)
+        ):
+            self.routes.append((method.upper(), src, self.mkredirect(dst)))
         # Serve api.js
         if self.js:
             self.routes.append(("GET", "/api.js", self.api_js))
