@@ -10,10 +10,10 @@ import aiohttp
 
 from dffml.model.slr import SLRModel
 from dffml.source.json import JSONSource
-from dffml import Record, Features, DefFeature, save, train, accuracy
+from dffml import Record, Features, Feature, save, train, accuracy
 from dffml.util.asynctestcase import AsyncTestCase
 
-from dffml_service_http.cli import HTTPService
+from dffml_service_http.cli import HTTPService, RedirectFormatError
 from dffml_service_http.util.testing import ServerRunner, ServerException
 
 from .test_routes import TestRoutesMultiComm
@@ -66,23 +66,6 @@ class TestCreateTLS(AsyncTestCase):
 
 
 class TestServer(AsyncTestCase):
-    @classmethod
-    def setUpClass(cls):
-        super().setUpClass()
-        cls._exit_stack = contextlib.ExitStack()
-        cls.exit_stack = cls._exit_stack.__enter__()
-        cls.exit_stack.enter_context(
-            patch(
-                "dffml.df.base.OperationImplementation.load",
-                new=TestRoutesMultiComm.patch_operation_implementation_load,
-            )
-        )
-
-    @classmethod
-    def tearDownClass(cls):
-        super().tearDownClass()
-        cls._exit_stack.__exit__(None, None, None)
-
     def url(self, cli):
         return f"http://{cli.addr}:{cli.port}"
 
@@ -135,6 +118,22 @@ class TestServer(AsyncTestCase):
                     )
                 )
 
+    async def test_portfile(self):
+        with tempfile.TemporaryDirectory() as tempdir:
+            portfile_path = pathlib.Path(tempdir, "portfile.int")
+            async with ServerRunner.patch(HTTPService.server) as tserver:
+                cli = await tserver.start(
+                    HTTPService.server.cli(
+                        "-insecure",
+                        "-port",
+                        "0",
+                        "-portfile",
+                        str(portfile_path),
+                    )
+                )
+                self.assertTrue(portfile_path.is_file())
+                self.assertEqual(cli.port, int(portfile_path.read_text()))
+
     async def test_mc_config(self):
         with tempfile.TemporaryDirectory() as tempdir:
             # URLs for endpoints
@@ -152,7 +151,7 @@ class TestServer(AsyncTestCase):
                 json.dumps(
                     {
                         "path": hello_world_url,
-                        "presentation": "json",
+                        "output_mode": "json",
                         "asynchronous": False,
                     },
                     sort_keys=True,
@@ -163,7 +162,7 @@ class TestServer(AsyncTestCase):
                 json.dumps(
                     {
                         "path": hello_blank_url,
-                        "presentation": "json",
+                        "output_mode": "json",
                         "asynchronous": False,
                     },
                     sort_keys=True,
@@ -229,8 +228,8 @@ class TestServer(AsyncTestCase):
         with tempfile.TemporaryDirectory() as tempdir:
             # Model the HTTP API will pre-load
             model = SLRModel(
-                features=Features(DefFeature("f1", float, 1)),
-                predict=DefFeature("ans", int, 1),
+                features=Features(Feature("f1", float, 1)),
+                predict=Feature("ans", int, 1),
                 directory=tempdir,
             )
 
@@ -284,7 +283,7 @@ class TestServer(AsyncTestCase):
         with tempfile.TemporaryDirectory() as tempdir:
             # Source the HTTP API will pre-load
             source = JSONSource(
-                filename=str(pathlib.Path(tempdir, "source.json")),
+                filename=pathlib.Path(tempdir, "source.json"),
                 allowempty=True,
                 readwrite=True,
             )
@@ -302,10 +301,51 @@ class TestServer(AsyncTestCase):
                         "-sources",
                         "mysource=json",
                         "-source-mysource-filename",
-                        source.config.filename,
+                        str(source.config.filename),
                     )
                 )
                 async with self.get(
                     cli, "/source/mysource/record/myrecord"
                 ) as r:
                     self.assertEqual(await r.json(), myrecord.export())
+
+    async def test_redirect_format_error(self):
+        with self.assertRaises(RedirectFormatError):
+            async with ServerRunner.patch(HTTPService.server) as tserver:
+                await tserver.start(
+                    # Missing METHOD
+                    HTTPService.server.cli(
+                        "-insecure",
+                        "-port",
+                        "0",
+                        "-redirect",
+                        "/",
+                        "/index.html",
+                    )
+                )
+
+    async def test_redirect(self):
+        with tempfile.TemporaryDirectory() as tempdir:
+            pathlib.Path(tempdir, "index.html").write_text("Hello World")
+            pathlib.Path(tempdir, "mysignup").write_text("MySignUp")
+            async with ServerRunner.patch(HTTPService.server) as tserver:
+                cli = await tserver.start(
+                    HTTPService.server.cli(
+                        "-insecure",
+                        "-port",
+                        "0",
+                        "-static",
+                        tempdir,
+                        "-redirect",
+                        "GET",
+                        "/",
+                        "/index.html",
+                        "GET",
+                        "/signup",
+                        "/mysignup",
+                    )
+                )
+                async with self.get(cli, "/") as r:
+                    self.assertEqual(await r.text(), "Hello World")
+                async with self.get(cli, "/signup") as r:
+                    self.assertEqual(await r.text(), "MySignUp")
