@@ -1,21 +1,22 @@
 import os
 import io
+import aiohttp
+import asyncio
 import pathlib
 import tempfile
 from unittest.mock import patch
 from contextlib import asynccontextmanager, ExitStack, AsyncExitStack
 from typing import AsyncIterator, Dict
 
-import aiohttp
-
+from dffml import DataFlow
 from dffml.base import config
 from dffml.record import Record
-from dffml.df.base import BaseConfig
+from dffml.df.base import BaseConfig, op
 from dffml.operation.output import GetSingle
 from dffml.util.entrypoint import EntrypointNotFound
 from dffml.model.model import ModelContext, Model
 from dffml.model.accuracy import Accuracy
-from dffml.feature import DefFeature
+from dffml.feature import Feature
 from dffml.source.source import Sources
 from dffml.source.csv import CSVSourceConfig
 from dffml.util.cli.arg import parse_unknown
@@ -28,6 +29,8 @@ from dffml_service_http.routes import (
     SOURCE_NOT_LOADED,
     MODEL_NOT_LOADED,
     MODEL_NO_SOURCES,
+    HTTPChannelConfig,
+    DISALLOW_CACHING,
 )
 from dffml_service_http.util.testing import (
     ServerRunner,
@@ -57,6 +60,27 @@ class TestRoutesService(TestRoutesRunning, AsyncTestCase):
             async with self.get("/non-existant"):
                 pass  # pramga: no cov
 
+    def set_no_cache_do_not_set_any_headers(self, response):
+        pass
+
+    async def test_check_allow_caching_header_not_found(self):
+        self.cli.set_no_cache = self.set_no_cache_do_not_set_any_headers
+        with self.assertRaisesRegex(Exception, "No cache header .* not in"):
+            async with self.get("/non-existant"):
+                pass  # pramga: no cov
+
+    def set_no_cache_bad_values_for_headers(self, response):
+        for header, value in DISALLOW_CACHING.items():
+            response.headers[header] = "BAD!"
+
+    async def test_check_allow_caching_header_not_correct(self):
+        self.cli.set_no_cache = self.set_no_cache_bad_values_for_headers
+        with self.assertRaisesRegex(
+            Exception, "No cache header .* should have been .* but was"
+        ):
+            async with self.get("/non-existant"):
+                pass  # pramga: no cov
+
 
 class TestRoutesServiceUpload(TestRoutesRunning, AsyncTestCase):
     async def test_success(self):
@@ -153,7 +177,7 @@ class TestRoutesList(TestRoutesRunning, AsyncTestCase):
 
 class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
     async def test_source(self):
-        config = parse_unknown(
+        config = await parse_unknown(
             "--source-filename", "dataset.csv", "-source-allowempty"
         )
         async with self.post("/configure/source/csv/salary", json=config) as r:
@@ -162,7 +186,7 @@ class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
             self.assertEqual(
                 self.cli.app["sources"]["salary"].config,
                 CSVSourceConfig(
-                    filename="dataset.csv",
+                    filename=pathlib.Path("dataset.csv"),
                     tag="untagged",
                     key="key",
                     tagcol="tag",
@@ -175,7 +199,7 @@ class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
                     self.assertIn("salaryctx", self.cli.app["source_contexts"])
 
     async def test_source_error(self):
-        config = parse_unknown("--source-file", "dataset.csv")
+        config = await parse_unknown("--source-file", "dataset.csv")
         with self.assertRaisesRegex(ServerException, "missing.*filename"):
             async with self.post("/configure/source/csv/salary", json=config):
                 pass  # pramga: no cov
@@ -191,7 +215,7 @@ class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
         with tempfile.TemporaryDirectory() as tempdir, patch.object(
             Model, "load", new=model_load
         ):
-            config = parse_unknown(
+            config = await parse_unknown(
                 "--model-directory",
                 tempdir,
                 "--model-features",
@@ -210,10 +234,10 @@ class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
                     FakeModelConfig(
                         directory=pathlib.Path(tempdir),
                         features=Features(
-                            DefFeature("Years", int, 1),
-                            DefFeature("Experiance", int, 1),
+                            Feature("Years", int, 1),
+                            Feature("Experiance", int, 1),
                         ),
-                        predict=DefFeature("Salary", float, 1),
+                        predict=Feature("Salary", float, 1),
                     ),
                 )
                 with self.subTest(context="salaryctx"):
@@ -228,7 +252,7 @@ class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
 
     async def test_model_config_error(self):
         # Should be directory, not folder
-        config = parse_unknown("--model-directory", "mymodel_dir")
+        config = await parse_unknown("--model-directory", "mymodel_dir")
         with patch.object(Model, "load", new=model_load):
             with self.assertRaisesRegex(ServerException, "missing.*features"):
                 async with self.post(
@@ -256,34 +280,6 @@ class TestRoutesConfigure(TestRoutesRunning, AsyncTestCase):
 
 
 class TestRoutesMultiComm(TestRoutesRunning, AsyncTestCase):
-    OPIMPS = {"formatter": formatter, "get_single": GetSingle, "remap": remap}
-
-    @classmethod
-    def patch_operation_implementation_load(cls, loading):
-        try:
-            return cls.OPIMPS[loading].imp
-        except KeyError as error:
-            raise EntrypointNotFound(
-                f"{loading} not found in {list(cls.OPIMPS.keys())}"
-            ) from error
-
-    @classmethod
-    def setUpClass(cls):
-        super().setUpClass()
-        cls._exit_stack = ExitStack()
-        cls.exit_stack = cls._exit_stack.__enter__()
-        cls.exit_stack.enter_context(
-            patch(
-                "dffml.df.base.OperationImplementation.load",
-                new=cls.patch_operation_implementation_load,
-            )
-        )
-
-    @classmethod
-    def tearDownClass(cls):
-        super().tearDownClass()
-        cls._exit_stack.__exit__(None, None, None)
-
     async def test_no_post(self):
         url: str = "/some/url"
         message: str = "Hello World"
@@ -296,7 +292,7 @@ class TestRoutesMultiComm(TestRoutesRunning, AsyncTestCase):
             f"/multicomm/self/register",
             json={
                 "path": url,
-                "presentation": "json",
+                "output_mode": "json",
                 "asynchronous": False,
                 "dataflow": HELLO_WORLD_DATAFLOW.export(),
             },
@@ -321,7 +317,7 @@ class TestRoutesMultiComm(TestRoutesRunning, AsyncTestCase):
             f"/multicomm/self/register",
             json={
                 "path": url,
-                "presentation": "json",
+                "output_mode": "json",
                 "asynchronous": False,
                 "dataflow": HELLO_BLANK_DATAFLOW.export(),
             },
@@ -343,6 +339,32 @@ class TestRoutesMultiComm(TestRoutesRunning, AsyncTestCase):
                 {"Feedface": {"response": message}}, await response.json()
             )
 
+    async def test_immediate_response(self):
+        url: str = "/some/url"
+        event = asyncio.Event()
+
+        @op()
+        async def my_event_setter(trigger: int) -> None:
+            event.set()
+
+        # Register the data flow
+        await self.cli.register(
+            HTTPChannelConfig(
+                path=url,
+                dataflow=DataFlow.auto(my_event_setter),
+                input_mode=f"text:{my_event_setter.op.inputs['trigger'].name}",
+                output_mode="json",
+                immediate_response={
+                    "status": 200,
+                    "content_type": "application/json",
+                    "data": {"immediate": "response"},
+                },
+            )
+        )
+        async with self.post(url, json="trigger") as response:
+            self.assertEqual(await response.json(), {"immediate": "response"})
+        await asyncio.wait_for(event.wait(), timeout=2)
+
 
 class TestRoutesSource(TestRoutesRunning, AsyncTestCase):
     async def setUp(self):
